# authentic2 - versatile identity manager
# Copyright (C) 2010-2022 Entr'ouvert
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from django.utils.translation import gettext_lazy as _

from authentic2.apps.journal.models import EventTypeDefinition
from authentic2.apps.journal.utils import form_to_old_new

from .models import SAMLAuthenticator


class SAMLAuthenticatorEvents(EventTypeDefinition):
    @classmethod
    def record(cls, *, user, session, related_object, data=None):
        data = data or {}
        data.update({'related_object': repr(related_object)})
        super().record(user=user, session=session, references=[related_object.authenticator], data=data)


class SAMLAuthenticatorRelatedObjectCreation(SAMLAuthenticatorEvents):
    name = 'authenticator.saml.related_object.creation'
    label = _('SAML authenticator related object creation')

    @classmethod
    def get_message(cls, event, context):
        (authenticator,) = event.get_typed_references(SAMLAuthenticator)
        related_object = event.get_data('related_object')
        if context != authenticator:
            return _('creation of {related_object} in authenticator "{authenticator}"').format(
                related_object=related_object, authenticator=authenticator
            )
        else:
            return _('creation of %s') % related_object


class SAMLAuthenticatorRelatedObjectEdit(SAMLAuthenticatorEvents):
    name = 'authenticator.saml.related_object.edit'
    label = _('SAML authenticator related object edit')

    @classmethod
    def record(cls, *, user, session, form):
        super().record(
            user=user,
            session=session,
            related_object=form.instance,
            data=form_to_old_new(form),
        )

    @classmethod
    def get_message(cls, event, context):
        (authenticator,) = event.get_typed_references(SAMLAuthenticator)
        related_object = event.get_data('related_object')
        new = event.get_data('new') or {}
        edited_attributes = ', '.join(new) or ''
        if context != authenticator:
            return _('edit {related_object} in authenticator "{authenticator}" ({change})').format(
                related_object=related_object,
                authenticator=authenticator,
                change=edited_attributes,
            )
        else:
            return _('edit {related_object} ({change})').format(
                related_object=related_object, change=edited_attributes
            )


class SAMLAuthenticatorRelatedObjectDeletion(SAMLAuthenticatorEvents):
    name = 'authenticator.saml.related_object.deletion'
    label = _('SAML authenticator related object deletion')

    @classmethod
    def get_message(cls, event, context):
        (authenticator,) = event.get_typed_references(SAMLAuthenticator)
        related_object = event.get_data('related_object')
        if context != authenticator:
            return _('deletion of {related_object} in authenticator "{authenticator}"').format(
                related_object=related_object, authenticator=authenticator
            )
        else:
            return _('deletion of %s') % related_object
