from django.apps import apps
from django.forms.models import modelform_factory
from django.http import Http404
from django.shortcuts import get_object_or_404, render
from django.template.loader import render_to_string
from django.urls import reverse
from django.views.generic import CreateView, DeleteView, UpdateView
from mellon.utils import get_idp

from authentic2.manager.views import MediaMixin, TitleMixin
from authentic2.utils.misc import redirect_to_login

from .forms import RoleChoiceField
from .models import (
    AddRoleAction,
    RenameAttributeAction,
    SAMLAttributeLookup,
    SAMLAuthenticator,
    SetAttributeAction,
)


def login(request, authenticator, *args, **kwargs):
    context = kwargs.pop('context', {}).copy()
    submit_name = 'login-saml-%s' % authenticator.slug
    context['submit_name'] = submit_name
    context['authenticator'] = authenticator
    if request.method == 'POST' and submit_name in request.POST:
        from .adapters import AuthenticAdapter

        settings = authenticator.settings
        AuthenticAdapter().load_idp(settings, authenticator.order)
        return redirect_to_login(
            request, login_url='mellon_login', params={'entityID': settings['ENTITY_ID']}
        )
    return render(
        request,
        ['authentic2_auth_saml/login_%s.html' % authenticator.slug, 'authentic2_auth_saml/login.html'],
        context,
    )


def profile(request, *args, **kwargs):
    context = kwargs.pop('context', {})
    user_saml_identifiers = request.user.saml_identifiers.all()
    if not user_saml_identifiers:
        return ''
    for user_saml_identifier in user_saml_identifiers:
        user_saml_identifier.idp = get_idp(user_saml_identifier.issuer.entity_id)
    context['user_saml_identifiers'] = user_saml_identifiers
    return render_to_string('authentic2_auth_saml/profile.html', context, request=request)


class SAMLAuthenticatorMixin(MediaMixin, TitleMixin):
    allowed_models = (SAMLAttributeLookup, RenameAttributeAction, SetAttributeAction, AddRoleAction)

    def dispatch(self, request, *args, **kwargs):
        self.authenticator = get_object_or_404(SAMLAuthenticator, pk=kwargs.get('authenticator_pk'))

        model_name = kwargs.get('model_name')
        if model_name not in (x._meta.model_name for x in self.allowed_models):
            raise Http404()
        self.model = apps.get_model('authentic2_auth_saml', model_name)

        return super().dispatch(request, *args, **kwargs)

    def get_form_class(self):
        return modelform_factory(
            self.model, exclude=('authenticator',), field_classes={'role': RoleChoiceField}
        )

    def get_form_kwargs(self):
        kwargs = super().get_form_kwargs()
        if not kwargs.get('instance'):
            kwargs['instance'] = self.model()
        kwargs['instance'].authenticator = self.authenticator
        return kwargs

    def get_success_url(self):
        return (
            reverse('a2-manager-authenticator-detail', kwargs={'pk': self.authenticator.pk})
            + '#open:%s' % self.model._meta.model_name
        )

    @property
    def title(self):
        return self.model._meta.verbose_name


class RelatedObjectAddView(SAMLAuthenticatorMixin, CreateView):
    template_name = 'authentic2/manager/form.html'

    def form_valid(self, form):
        resp = super().form_valid(form)
        self.request.journal.record(
            'authenticator.saml.related_object.creation', related_object=form.instance
        )
        return resp


add_related_object = RelatedObjectAddView.as_view()


class RelatedObjectEditView(SAMLAuthenticatorMixin, UpdateView):
    template_name = 'authentic2/manager/form.html'

    def form_valid(self, form):
        resp = super().form_valid(form)
        self.request.journal.record('authenticator.saml.related_object.edit', form=form)
        return resp


edit_related_object = RelatedObjectEditView.as_view()


class RelatedObjectDeleteView(SAMLAuthenticatorMixin, DeleteView):
    template_name = 'authentic2/authenticators/authenticator_delete_form.html'
    title = ''

    def delete(self, *args, **kwargs):
        self.request.journal.record(
            'authenticator.saml.related_object.deletion', related_object=self.get_object()
        )
        return super().delete(*args, **kwargs)


delete_related_object = RelatedObjectDeleteView.as_view()
